# This file is part of 'PyAsync'
# Copyright 2000 Constantinos A. Kotsokalis <ckotso@cs.ntua.gr>
#
# PyAsync is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyAsync is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyAsync; see the file COPYING. If not, write to the
# Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.


# print() statements throughout the file are for debugging reasons, do
# not comment them out.

import socket
import select
import name_services
import time

# Check if IPPROTO_IP is defined (it is in Python 1.5.2,
# but not in earlier versions)
if 'IPPROTO_IP' in dir(socket):
	IPPROTO_IP = socket.IPPROTO_IP
else:
	IPPROTO_IP = 0

# Asynchronous network engine class
class network_engine:

	# Object initialization: Create and bind listening socket
	def __init__(self, cfg, errobj):

		self.cfg = cfg

		if not cfg.has_key('BIND_PORT'):
			errobj.eprint_exit("No bind port in configuration dictionary")
		else:
			port = cfg['BIND_PORT']

		# Used to keep track of connected sockets
		self.inlist = []

		self.last_check = time.time()
		self.dnsobj = name_services.dnscache(errobj)

		self.errobj = errobj

		# Create listening socket
		try:
			ls = socket.socket(socket.AF_INET, socket.SOCK_STREAM, IPPROTO_IP)
		except socket.error, msg:
			errobj.eprint_exit("socket: ", msg)

		# Set it to ignore TIME_WAITs before binding
		try:
			ls.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		except socket.error, msg:
			errobj.eprint("setsockopt: ", msg)

		addr = '', port
		
		# Bind it
		try:
			ls.bind(addr)
		except socket.error, msg:
			errobj.eprint_exit("bind: ", msg)

		# Initialize listening queue with a size of 5
		try:
			ls.listen(5)
		except socket.error, msg:
			errobj.eprint_exit("listen: ", msg)

		# Append it to the list of active sockets
		self.inlist.append(ls)


	# Used by the application whenever it needs to connect to a remote
	# host and then wait for data from it. It is needed in order to
	# add the new socket to the list of active sockets, so that it
	# can pass read data to the application later on.
	def new_sock_client(self, host, port):

		errobj = self.errobj

		# Create the socket, or return None on failure
		try:
			cli = socket.socket(socket.AF_INET, socket.SOCK_STREAM, IPPROTO_IP)
			# print "New socket created: " + `cli.fileno()`
		except socket.error, msg:
			errobj.eprint("socket: ", msg)
			return None, msg

		# Connect to the remote server & port, or return None on failure
		try:
			cli.connect(self.dnsobj.name_check(host), port)
		except socket.error, msg:
			errobj.eprint("connect: ", msg)
			# print "An error occured on connect(), closing down: " + `cli.fileno()`
			cli.close()
			return None, msg

		# Append tp the list of active connections
		self.inlist.append(cli)

		# Return socket object to the application
		return cli, None


	# Used to write data to an active socket.
	def sock_write(self, sock, buff):
		inlist = self.inlist

		if not (sock in inlist):
			return None

		try:
			# print "Will try to write to: " + `sock.fileno()`
			buff_size = sock.send(buff)
			return buff_size
		except socket.error, msg:
			# print "An error occured while trying to write to: " + `sock.fileno()`
			self.errobj.eprint("send: ", msg)
			return None


	# Used to close sockets gracefully. onClose() is called here too!!!
	def sock_close(self, sock):
		inlist = self.inlist

		if not (sock in inlist):
			return None

		try:
			inlist.remove(sock)
			# print "Closing " + `sock.fileno()`
			sock.close()
			return 1
		except socket.error, msg:
			self.errobj.eprint("close: ", msg)
			return None


	# Engine main loop: Accepts new connections, reads from existing
	# connections when there is data available, and executes functions
	# passed as arguments on certain conditions (see following comments)
	def loop(self, onAccept, onRead, onClose, onLoopEnd):

		errobj = self.errobj
		inlist = self.inlist
		self.onClose = onClose

		# Go into infinite loop. Stops with Ctrl-C (interrupt)
	 	while (1):

			# Check on existing active connections to see whether there
			# is data available. Data on listening sockets means there
			# is a new connection request.
			try:
				rin = select.select(inlist, [], [])[0]
			except select.error, msg:
				errobj.eprint_exit("select: ", msg)

			# For all sockets that have data available...
			for sock in rin:

				if not sock in inlist:
					continue

				# Listening socket: There's an inbound connection. Accept it,
				# execute the onAccept function argument, and append to the
				# list of active connections. The new socket object is
				# passed as an argument to the onAccept user-defined function
				if sock == inlist[0]:
					try:
						newconn = sock.accept()
						# print "Accepted " + `newconn.fileno()`
						inlist.append(newconn[0])
						onAccept(newconn[0], newconn[1])
					except socket.error, msg:
						errobj.eprint("accept: ", msg)
					continue

				# Read data from sockets apart from the listening one,
				# and execute the onRead function with the socket we
				# read data from and the buffer itself. If we read
				# a buffer with a size of zero, that means the other
				# end has closed the connection. In this case, we
				# execute the onClose function.
				try:
					# print "Will try to read from: " + `sock.fileno()`
					buff = sock.recv(8192)
				except socket.error, msg:
					# print "An error occured while trying to read from: " + `sock.fileno()`
					errobj.eprint("recv: ", msg)
					continue

				if not len(buff):
					inlist.remove(sock)
					onClose(sock)
					# print "Closing " + `sock.fileno()`
					sock.close()
				else:
					onRead(sock, buff)

			# This is used for routine checks after data is read from
			# all sockets, on every cycle executed.
			onLoopEnd()

			# Update DNS dictionary if needed
			now = time.time()
			if self.cfg['DNS_CHECK_SEC'] > 0:
				if now - self.last_check >= self.cfg['DNS_CHECK_SEC']:
					retval = self.dnsobj.dns_update()
					self.last_check = now


